from __future__ import absolute_import
from gaw.jsonsocketserver.client import JsonSocketClient
import time

ip = '127.0.0.1'
port = 4444

client = JsonSocketClient()

print(client.request(ip, port, 'plus', dict(a=10.5,
                                            b=20)))

times = 10000
start_time = time.time()
for i in range(times):
    client.request(ip, port, 'home', '')
end_time = time.time()


print('taken:', (end_time - start_time) * 1000 / times)


from __future__ import absolute_import
from gaw.postoffice.client import PostofficeClient

client = PostofficeClient(ip='127.0.0.1', port=4444, verbose=True)
response = client.send(dict(a=10))

print('response:', response)


from __future__ import print_function, absolute_import
from gaw.jsonsocketserver.datatype import RequestDataType, ResponseDataType
from gaw.postoffice import PostofficeClient
from threading import Thread
import uuid
import datetime
import time
import eventlet

class JsonSocketClient:

    def __init__(self, client_lifetime=30, verbose=False):
        self.client_lifetime = client_lifetime
        self.verbose = verbose

        self._client_pool = dict()
        self._is_busy = dict()
        self._last_act_at = dict()

        self._cleaner = Thread(target=self._cleaner)
        self._cleaner.daemon = True
        self._cleaner.start()

    def request(self, ip, port, path, payload):
        # may retry many times
        while True:
            client = self._get_client(ip, port)
            assert isinstance(client, PostofficeClient)

            if self.verbose:
                print('jsonsocketclient: requesting ip', ip, 'port', port, 'path', path, 'payload', payload)

            request = RequestDataType(id=uuid.uuid1().int,
                                      path=path,
                                      payload=payload)

            try:
                raw_response = client.send(request.dict())
            except Exception as err:
                print('jsonsocketclient: connection error retrying ...')
                # delete the old client
                self._delete(self._key(ip, port))
                eventlet.sleep(1)
                # retry
                continue

            if self.verbose:
                print('jsonsocketclient: response ', raw_response)

            response = ResponseDataType.parse(raw_response)

            if response.resp_to != request.id:
                raise ValueError('request error: wrong response id')

            if not response.success:
                exception = response.payload
                type = exception['type']
                name = exception['name']
                message = exception['message']
                trace = exception['trace']
                raise Exception('\ntype: {}\nname: {}\nmessage: {}\ntrace: {}'.format(type, name, message, trace))

            return response.payload

    # private
    def _key(self, ip, port):
        return '{}:{}'.format(ip, port)

    def _init_client(self, ip, port):
        while True:
            try:
                return PostofficeClient(ip, port, verbose=self.verbose)
            except Exception:
                print('jsonsocketclient: host is down retrying ...')
                eventlet.sleep(5)

    def _get_client(self, ip, port):
        key = self._key(ip, port)

        if key not in self._client_pool:
            self._client_pool[key] = self._init_client(ip, port)

        self._last_act_at[key] = datetime.datetime.now()
        self._is_busy[key] = True

        return self._client_pool[key]

    def _return_client(self, ip, port):
        key = self._key(ip, port)
        self._last_act_at[key] = datetime.datetime.now()
        self._is_busy[key] = False

    def _delete(self, key):
        del self._client_pool[key]
        del self._last_act_at[key]
        del self._is_busy[key]

        if self.verbose:
            print('jsonsocketclient: cleared', key)

    def _cleaner(self):
        while True:
            time.sleep(10) # cleaning interval
            now = datetime.datetime.now()

            for key in list(self._client_pool.keys()):
                if self._is_busy[key]:
                    continue

                last_act = self._last_act_at[key]

                if now - last_act > datetime.timedelta(seconds=self.client_lifetime):
                    self._delete(key)


import json

class ConnectionTerminated(Exception):
    def __init__(self): super(ConnectionTerminated, self).__init__('Connection Terminated')

def send(socket, data):
    # serialize
    try:
        serialized = json.dumps(data)
    except (TypeError, ValueError) as e:
        raise Exception('You can only send JSON-serializable data')

    # send the data
    socket.send(bytearray('{}\n'.format(len(serialized)), 'ascii'))
    socket.sendall(bytearray(serialized, 'utf-8'))

def recieve(socket):
    # read the length of the data, letter by letter until we reach EOL
    length_str = b''
    char = socket.recv(1)

    # check termination
    if not char:
        raise ConnectionTerminated()

    # got a message
    # read message length
    while char != b'\n':
        length_str += char
        char = socket.recv(1)

    payload_length = int(length_str)

    # get the payload
    buffer = memoryview(bytearray(payload_length))
    next_offset = 0
    while payload_length - next_offset > 0:
        recv_size = socket.recv_into(buffer[next_offset:], payload_length - next_offset)
        next_offset += recv_size

    # deserialize payload
    try:
        data = json.loads(buffer.tobytes().decode('utf-8'))
    except (TypeError, ValueError) as e:
        raise Exception('Data received was not in JSON format')

    return data


from __future__ import print_function, absolute_import
from eventlet.green import socket
from gaw.postoffice.core import send, recieve

class PostofficeClient:

    def __init__(self, ip, port, verbose=False):
        self.ip = ip
        self.port = port
        self.verbose = verbose

        self.socket = socket.socket()
        self.socket.connect((ip, port))

    def send(self, data):
        if self.verbose:
            print('postofficeclient: sending data', data)

        send(self.socket, data)

        if self.verbose:
            print('postofficeclient: sending done')

        response = recieve(self.socket)

        if self.verbose:
            print('postofficeclient: receciving done')

        return response


from __future__ import print_function, absolute_import
import eventlet
from gaw.postoffice.core import recieve, send, ConnectionTerminated

class PostofficeServer:

    def __init__(self, ip, port, on_message, verbose=False):
        print('postoffice: listening on ip:', ip, 'port:', port)

        self.verbose = verbose
        self.on_message = on_message

        server = eventlet.listen((ip, port))
        pool = eventlet.GreenPool()
        while True:
            try:
                socket, address = server.accept()
                pool.spawn_n(self._on_connection, socket, address)
            except (SystemExit, KeyboardInterrupt):
                break

    def _on_connection(self, socket, address):
        if self.verbose:
            print('postoffice: get a connection from', address)

        # loop until connection terminated
        while True:

            try:
                data = recieve(socket)
                print('postoffice: get a message ', str(data)[:150], '...')
            except ConnectionTerminated:
                if self.verbose:
                    print('postoffice: connection ended')
                return

            response = self.on_message(data)
            send(socket, response)